"""
Functions for working with configuration file selection and reading.

The :func:`which` convenience function returns the first file that should be used as configuration file from a list.
"""

import os
import typing as t

from boltons.iterutils import first

__all__ = [
    "which",
]


def which(
    cfgfile: str = None,
    candidates: t.List[str] = None,
    default: str = "/dev/null",
    verify: t.Callable = os.path.exists,
) -> t.Optional[str]:
    """
    Convenience function to return the configuration file to read.

    Parameters
    ----------
    cfgfile: str, None
        a single path. If not None, then it will be returned without calling the verify function.

    candidates: t.List[str], None
        a list of paths. If cfgfile is None, then the verify function will be
        called for each of these paths and the first one for which verify
        returns True will be returned.

        If no file is matched, then :file:`/dev/null` will be returned.

    verify: t.Callable
        callable that will be called for every file in candidate. Defaults to os.path.exists.

    default: str, None
        value to be returned if cfgfile is None and no match from candidates was found

    Returns
    -------
    str:
        cfgfile, a value from candidates, or default

    Example
    -------

    .. code-block:: python

        import argparse, sys
        from torxtools import cfgtools, pathtools

        defaults = pathtools.expandpath(["~/.my-cfgfile"], ["/etc/cfgfile"])
        parser = argparse.ArgumentParser()
        parser.add_argument('--cfgfile')

        args = parser.parse_args(sys.argv[:1])
        print(cfgtools.which(args.cfgfile, defaults))

    """
    if not callable(verify):
        raise TypeError(f"expected verify to be a callable, not: {verify}")

    if cfgfile:
        return cfgfile
    return first(candidates or [], default=default, key=verify)



"""
Dictionary handling convenience functions.
"""

from copy import deepcopy

from boltons import iterutils


def strip_none(_p, k, v):
    return k is not None and v is not None


def _deepmerge(src, dest, path=None):
    """
    Take every key, value from src and merge it recursively into dest.
    Adapted from https://stackoverflow.com/questions/7204805

    :param dest: destination dict
    :param src: source dict
    :returns: merged dict
    """
    if path is None:
        path = []

    for key in src:
        if key not in dest:
            dest[key] = src[key]
        elif isinstance(dest[key], dict) and isinstance(src[key], dict):
            _deepmerge(src[key], dest[key], path + [str(key)])
        elif dest[key] == src[key]:
            pass  # same leaf value
        else:
            dest[key] = src[key]
    return dest


def deepmerge(*args, visit=None):
    """
    Deprecated / Warning: Quality not met

    Take every key and value from first and merge it recursively into second
    without changing first nor second.

    :param args: sources, left to right
    :returns: A merged dictionary
    """
    if not args:
        return None

    context = {}
    for i in range(len(args)):  # pylint: disable=consider-using-enumerate
        if args[i] is None:
            source = {}
        else:
            source = deepcopy(args[i])

        if visit:
            source = iterutils.remap(source, visit=visit)

        context = _deepmerge(source, context)

    return context



"""
Parser types for command-line options, arguments and sub-commands

Example
-------

.. code-block::

    parser.add_argument(
        "--size", "-s",
        dest="size",
        help="[MB] Minimal size of attachment",
        type=torxtools.argparse.is_int_positive,
        default=100,
    )

"""

from argparse import ArgumentTypeError

__all__ = [
    "is_int_positive",
    "is_int_positive_or_zero",
    "is_int_negative",
    "is_int_negative_or_zero",
]


def is_int_positive(value: int) -> int:
    """
    Verify that argument passed is a positive integer.
    """
    number = int(value)
    if number <= 0:
        raise ArgumentTypeError(f"value '{value}' must be positive")
    return number


def is_int_positive_or_zero(value: int) -> int:
    """
    Verify that argument passed is a positive integer or zero.
    """
    number = int(value)
    if number < 0:
        raise ArgumentTypeError(f"value '{value}' must be positive or zero")
    return number


def is_int_negative(value: int) -> int:
    """
    Verify that argument passed is a negative integer.
    """
    number = int(value)
    if number >= 0:
        raise ArgumentTypeError(f"value '{value}' must be negative")
    return number


def is_int_negative_or_zero(value: int) -> int:
    """
    Verify that argument passed is a negative integer or zero.
    """
    number = int(value)
    if number > 0:
        raise ArgumentTypeError(f"value '{value}' must be negative or zero")
    return number



__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]

__title__ = "torxtools"
__summary__ = "Less generic functionalities than Phillips and Pozidriv tools."
__uri__ = "https://gitlab.com/jlecomte/projects/python/torxtools"

__version__ = "0.14.2"

__author__ = "Julien Lecomte"
__email__ = "julien@lecomte.at"

__license__ = "MIT"
__copyright__ = f"2020-2022 {__author__}"



"""
Utilities for with-statement contexts
"""
# pylint: disable=invalid-name
#
import contextlib


class suppress(contextlib.suppress, contextlib.ContextDecorator):
    """
    A version of contextlib.suppress with decorator support.
    """


class reraise_from_none(contextlib.suppress, contextlib.ContextDecorator):
    """
    Similar to contextlib.suppress, but with decorator support, and that
    re-raises exception from None instead of hiding it.
    """

    def __exit__(self, exctype, excinst, _exctb):
        if exctype is None:
            return
        if issubclass(exctype, self._exceptions):
            raise excinst from None



#!/usr/bin/env python3
# template: https://gitlab.com/jlecomte/projects/python/reference-files

from setuptools import setup

from torxtools import __about__ as about

with open("requirements.txt", encoding="UTF-8") as fd:
    requirements = fd.read().splitlines()
    requirements = [x for x in requirements if x and not x.startswith("--")]

setup(
    name=about.__title__,
    version=about.__version__,
    author=about.__author__,
    author_email=about.__email__,
    url=about.__uri__,
    description=about.__summary__,
    license=about.__license__,
    install_requires=requirements,
)



"""
Functions for working with miscellaneous
"""
import os
import sys
import typing as t

__all__ = [
    "maincli",
    "get_package_about",
    "get_package_requirements",
]


def maincli(fn: t.Callable, *args, **kwargs) -> None:
    """
    TODO
    """
    try:
        return fn(*args, **kwargs)
    except KeyboardInterrupt:
        # exit code could be success or error, it all depends on if it's the
        # normal way of quitting the app.
        pass
    except SystemExit as err:
        if isinstance(err.code, int):
            return err.code
        print(err, file=sys.stderr)
        return 1
    except Exception as err:  # pylint: disable=broad-except
        print(f"error: {err}", file=sys.stderr)
    return 1


def get_package_requirements() -> t.List:
    """
    TODO
    """
    with open("requirements.txt", encoding="UTF-8") as fd:
        requirements = fd.read().splitlines()
        requirements = [x for x in requirements if x and not x.startswith("--")]
        return requirements


def get_package_about(path=".") -> t.List:
    """
    TODO
    """
    about = {}

    for file in os.listdir(path):
        file = path + "/" + file
        if not os.path.isdir(file):
            continue
        if not os.path.exists(file + "/__about__.py"):
            continue
        with open(file + "/__about__.py", encoding="UTF-8") as fd:
            exec(fd.read(), None, about)  # pylint: disable=exec-used
        break

    about = {k: v for k, v in about.items() if k.startswith("__")}
    return about
